"""pylaundry exceptions."""


class MessagePackerError(Exception):
    """Problem with encoding/decoding or encrypting/decrypting."""


class ResponseFormatError(Exception):
    """API response in wrong format."""


class CommunicationError(Exception):
    """Could not reach server."""


class Rejected(Exception):
    """Server understood but rejected the request."""


class UnexpectedError(Exception):
    """Something bad happened that we didn't expect."""


class NotLoggedIn(Exception):
    """User attempted to perform an action before logging in."""


class AuthenticationError(Exception):
    """User submitted invalid credentials."""


class VendFailure(Exception):
    """Failure vending."""


class MachineNotFound(Exception):
    """Machine not found."""


class VendLogFailure(Exception):
    """Failure logging vend."""



"""Global fixture functions."""

# pylint: disable = redefined-outer-name

from collections.abc import AsyncGenerator
from collections.abc import Generator

import aiohttp
from aioresponses import aioresponses
from pylaundry import Laundry
from pylaundry.const import API_ENDPOINT_URL
import pytest

from .http_bodies import get_http_body


#
# Meta Fixtures
#


@pytest.fixture  # type: ignore
def response_mocker() -> Generator:
    """Yield aioresponses."""
    with aioresponses() as mocker:
        yield mocker


@pytest.fixture  # type: ignore
@pytest.mark.asyncio  # type: ignore
async def laundry() -> AsyncGenerator:
    """Build and return dummy controller for testing without Alarm.com API."""

    async with aiohttp.ClientSession() as websession:
        yield Laundry(websession=websession)


#
# Response Fixtures
#


@pytest.fixture  # type: ignore
def additional_information__response__success(response_mocker: aioresponses) -> None:
    """Success response for additional information."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("additional_information__response__success"),
    )


@pytest.fixture  # type: ignore
def authentication__response__success(response_mocker: aioresponses) -> None:
    """Success response for authentication."""

    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("authentication__response__success"),
        headers={"CP_AUTH_TOKEN": "e94eca12-854f-409e-b32f-302805ed12d9"},
    )


@pytest.fixture  # type: ignore
def authentication__response__incorrect_credentials(
    response_mocker: aioresponses,
) -> None:
    """Incorrect credentials response for authentication."""

    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("authentication__response__incorrect_credentials"),
        headers={"CP_AUTH_TOKEN": "e94eca12-854f-409e-b32f-302805ed12d9"},
    )


@pytest.fixture  # type: ignore
def consolidated_refresh__response__success(response_mocker: aioresponses) -> None:
    """Success response for consolidated refresh."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("consolidated_refresh__response__success"),
    )


@pytest.fixture  # type: ignore
def general__response__incorrect_packing(response_mocker: aioresponses) -> None:
    """Incorrect packing response for any request."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("general__response__incorrect_packing"),
    )


@pytest.fixture  # type: ignore
def get_vend_price__response__success(response_mocker: aioresponses) -> None:
    """Success response for get vend price."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("get_vend_price__response__success"),
    )


@pytest.fixture  # type: ignore
def vend_log_topoff__response__success(response_mocker: aioresponses) -> None:
    """Success response for vend log."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("vend_log_topoff__response__success"),
    )


@pytest.fixture  # type: ignore
def virtual_vend_topoff__response__success(response_mocker: aioresponses) -> None:
    """Success response for virtual vend topoff."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("virtual_vend_topoff__response__success"),
    )


#
# Request Fixtures
#

# TODO: Test request objects.


@pytest.fixture  # type: ignore
def additional_information__request(response_mocker: aioresponses) -> None:
    """Request for additional information."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("additional_information__request"),
    )


@pytest.fixture  # type: ignore
def authentication__request(response_mocker: aioresponses) -> None:
    """Request for authentication."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("authentication__request"),
    )


@pytest.fixture  # type: ignore
def consolidated_refresh__request(response_mocker: aioresponses) -> None:
    """Request for consolidated refresh."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("consolidated_refresh__request"),
    )


@pytest.fixture  # type: ignore
def get_vend_price__request(response_mocker: aioresponses) -> None:
    """Request for get vend price."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("get_vend_price__request"),
    )


@pytest.fixture  # type: ignore
def vend_log__request(response_mocker: aioresponses) -> None:
    """Request for vend log."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("vend_log__request"),
    )


@pytest.fixture  # type: ignore
def virtual_vend__request(response_mocker: aioresponses) -> None:
    """Request for virtual vend."""

    # Refresh Response
    response_mocker.post(
        url=API_ENDPOINT_URL,
        status=200,
        body=get_http_body("virtual_vend__request"),
    )



"""pylaundry library."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from datetime import datetime
from datetime import timezone
from enum import Enum
import hashlib
import json
import logging
import uuid

import aiohttp
import dateutil.parser

from .const import API_ENDPOINT_URL
from .const import APPKEY
from .const import AUTH_TOKEN_KEY
from .const import EMPTY_AUTH_TOKEN
from .const import LOG_LEVEL_TRACE
from .const import REFRESH_REQUEST_PREHASH_SUFFIX
from .const import RESULT_CODE_KEY
from .const import RESULT_TEXT_KEY
from .const import ServerResponseCodes
from .exceptions import AuthenticationError
from .exceptions import CommunicationError
from .exceptions import MachineNotFound
from .exceptions import NotLoggedIn
from .exceptions import Rejected
from .exceptions import ResponseFormatError
from .exceptions import UnexpectedError
from .exceptions import VendFailure
from .exceptions import VendLogFailure
from .helpers import MessagePacker

__version__ = "v0.1.3"

log = logging.getLogger(__name__)


class MachineType(Enum):
    """Laundry machine types."""

    WASHER = "Washer"
    DRYER = "Dryer"
    UNKNOWN = "Unknown"


@dataclass
class LaundryMachine:
    """Representation of a washer or dryer."""

    id_: str
    type: MachineType
    number: str
    busy: bool | None
    minutes_remaining: int | None
    base_price: float | None
    topoff_price: float | None
    online: bool | None
    reader_serial: str | None


@dataclass
class LaundryProfile:
    """Metadata for location and user's virtual card."""

    location_address: str
    card_balance: float
    user_id: str
    user_token: str
    location_id: str
    database_id: str | None
    card_serial: str | None


class Laundry:
    """pylaundry's controller."""

    profile: LaundryProfile
    machines: dict
    encryption_keys: list[str]

    def __init__(self, websession: aiohttp.ClientSession) -> None:
        """Initialize pylaundry."""

        self._websession: aiohttp.ClientSession = websession
        self._first_request_id: str | None = None
        self._auth_token: str = EMPTY_AUTH_TOKEN

        self._username: str | None = None
        self._password: str | None = None

        self.installation_token = str(uuid.uuid4())

    async def async_login(self, username: str, password: str) -> None:
        """Log in to Laundry Link."""

        # We need to store these for error recovery in _send_request().
        self._username = username
        self._password = password

        try:
            request_data = [
                "Authenticate2",
                APPKEY,
                username,
                password,
                self.installation_token,
                '{"droidDisplay":"LMY47E","droidBrand":"Android","droidProduct":"sdk_phone_x86","droidDevice":"shamu","droidManufacturer":"motorola","droidModel":"Nexus'
                ' 6","droidHardware":"ranchu","droidSDK":28,"droidVersionRelease":"9","droidVersionIncremental":"4923214","droidVersionCodeName":"REL","droidIsRooted":false,"droidAppVersion":"4.09","droidAppBundleID":"com.esd.laundrylink.hercules"}',
            ]

            response = await self._send_request(json.dumps(request_data))

        except (
            CommunicationError,
            ResponseFormatError,
            Rejected,
            AuthenticationError,
        ) as err:
            raise err

        self._process_machine_data(
            response.get("Bundle", {}).get("MachinesInformation", {})
        )

        # Assemble user token
        user_token_raw = (
            f"{(user_id := response['UserID'])}{REFRESH_REQUEST_PREHASH_SUFFIX}"
        )

        log.log(LOG_LEVEL_TRACE, "Secret Raw:\n%s\n\n", user_token_raw)

        userhash_md5 = hashlib.md5(bytes(user_token_raw, "utf-8"))  # nosec

        user_token = userhash_md5.hexdigest()

        # Assemble profile
        self.profile = LaundryProfile(
            location_address=response["LocationAddress"],
            card_balance=response.get("Bundle", {})
            .get("CardInformation", {})
            .get("Balance"),
            user_id=user_id,
            location_id=response["LocationID"],
            database_id=response["DatabaseID"],
            card_serial=response.get("Bundle", {})
            .get("CardInformation", {})
            .get("AccountNumber"),
            user_token=user_token,
        )

    async def async_get_encryption_keys(self) -> None:
        """Get encryption keys from server."""

        # Purpose of these keys is TBD. Presumably, they're used for dispense / reload transactions.

        if self._auth_token == EMPTY_AUTH_TOKEN:
            raise NotLoggedIn

        request_data = ["GetAdditionalInformation", APPKEY]

        response = await self._send_request(json.dumps(request_data))

        if len(values := response.get("Values", [])) > 0 and isinstance(values, list):
            self.encryption_keys = values
        else:
            log.error("Failed to retrieve encryption keys.")

    async def async_refresh(self) -> None:
        """Get updated machine status."""

        if self._auth_token == EMPTY_AUTH_TOKEN:
            raise NotLoggedIn

        request_data = [
            "ConsolidatedRefresh",
            self.profile.user_token,
            self.profile.user_id,
        ]

        response = await self._send_request(json.dumps(request_data))

        # Refresh card balance.
        self.profile.card_balance = (
            balance
            if (balance := response.get("CardInformation", {}).get("Balance"))
            else None
        )

        # Refresh machine status.
        self._process_machine_data(response.get("MachinesInformation", {}))

    async def async_get_topoff_price(self, machine_id: str) -> None:
        """Get topoff price for single machine, then update machine with price."""

        if self._auth_token == EMPTY_AUTH_TOKEN:
            raise NotLoggedIn

        machine: LaundryMachine = self.machines[machine_id]

        request_data = [
            "GetVendPrice",
            self.profile.user_token,
            self.profile.database_id,
            machine.reader_serial,
        ]

        response = await self._send_request(json.dumps(request_data))

        if isinstance(machine := self.machines[machine_id], LaundryMachine):
            machine.topoff_price = response.get("TopoffPrice")

    async def _async_log_vend(
        self, machine_id: str, error_code: int, vend_success: bool
    ) -> None:
        """Log vend for a single machine."""

        # CyclePay logs each vend using this request, but this doesn't seem to be necessary to adding value to a machine.
        # Just using async_vend() is sufficient to vend and update the user-facing activity log.

        if self._auth_token == EMPTY_AUTH_TOKEN:
            raise NotLoggedIn

        try:
            machine = self.machines[machine_id]
        except AttributeError as err:
            raise MachineNotFound(machine_id) from err

        if not isinstance(machine, LaundryMachine):
            raise MachineNotFound(machine_id)

        request_data = [
            "CreateVendLogEntry",
            self.profile.user_token,
            self.profile.user_id,
            str(datetime.now(timezone.utc)),
            self.profile.card_serial,
            machine.reader_serial,
            machine.number,
            error_code,
            vend_success,
            False,
            machine.base_price,  # Always base price, even when topping off
        ]

        response = await self._send_request(json.dumps(request_data))

        if response.get("ResultCode") != 1:
            log.error("Failed to log vend. Response: %s", response)
            raise VendLogFailure

    async def async_vend(self, machine_id: str) -> None:
        """Vend a single machine and log result."""

        if self._auth_token == EMPTY_AUTH_TOKEN:
            raise NotLoggedIn

        machine: LaundryMachine = self.machines[machine_id]

        request_data = [
            "VirtualVend",
            self.profile.user_token,
            self.profile.database_id,
            machine.reader_serial,
            self.profile.card_serial,
        ]

        try:
            response = await self._send_request(json.dumps(request_data))
        except (
            UnexpectedError,
            ResponseFormatError,
            Rejected,
            CommunicationError,
        ) as err:
            log.error("Communication error while vending.")
            raise VendFailure from err

        if response.get("ResultCode") not in [1, 161]:
            log.error("Failed to vend machine. Response: %s", response)
            raise VendFailure

        log.debug("Vend successful.")

        # Bypassing log. See note in _async_log_vend() for details.

        # try:
        #     await self._async_log_vend(
        #         machine_id=machine_id,
        #         error_code=response["ResultCode"],
        #         vend_success=True,
        #     )
        # except (
        #     UnexpectedError,
        #     ResponseFormatError,
        #     Rejected,
        #     CommunicationError,
        #     VendLogFailure,
        # ) as err:
        #     log.error("Error logging vend.")
        #     raise VendLogFailure from err

    def _process_machine_data(self, machines_info_object: dict) -> None:
        """Update machine data from API MachinesInformation object."""

        if machines_info_object.get(RESULT_CODE_KEY) != 1:
            log.error("Problem with machines response: %s", machines_info_object)

        machines: dict = {}

        machine: dict
        for machine in machines_info_object.get("Machines", []):

            try:
                machine_id = machine["ReaderID"]

                machine_type = MachineType.UNKNOWN
                if (setup_type := machine.get("SetupType")) == "Dryer":
                    machine_type = MachineType.DRYER
                elif setup_type == "Washer":
                    machine_type = MachineType.WASHER

                # Determine how long ago state was reported.
                state_age_min = (
                    datetime.now(timezone.utc)
                    - dateutil.parser.isoparse(machine["StateDateTimeUtc"])
                ).total_seconds() / 60

                # Adjust minutes remaining by state age.
                minutes_remaining = max(
                    0, round(machine.get("MinutesRemaining", 0) - state_age_min)
                )

                machines[machine_id] = LaundryMachine(
                    id_=machine["ReaderID"],
                    type=machine_type,
                    number=machine["Label"],
                    busy=minutes_remaining > 0,
                    minutes_remaining=minutes_remaining,
                    base_price=machine.get("BasePrice"),
                    online=bool(is_online)
                    if (is_online := machine.get("IsOnline")) in [True, False]
                    else None,
                    reader_serial=machine.get("SerialNumber"),
                    topoff_price=None,
                )

            except KeyError:
                log.error("Failed to retrieve data for a machine: %s", machine)

        self.machines = machines

    async def _send_request(self, request_json: str, no_retry: bool = False) -> dict:
        """Send submitted request body to server. Handles body formatting and headers and updates session objects."""

        request_id, packed_request_data = MessagePacker.pack_client_request(
            request_body=request_json,
            first_request_id=self._first_request_id,
        )

        if self._first_request_id is None:
            self._first_request_id = request_id

        request_headers = {
            "CP_REQ_ID": request_id,
            AUTH_TOKEN_KEY: self._auth_token,
            "Content-Type": "application/x-www-form-urlencoded",
        }

        log.log(
            LOG_LEVEL_TRACE, "==============[ BUILDING REQUEST BEGIN ]=============="
        )
        log.log(LOG_LEVEL_TRACE, "** REQUEST HEADERS **")
        log.log(LOG_LEVEL_TRACE, request_headers)

        request_body = f"CP_REQ_DATA={packed_request_data}"

        log.log(LOG_LEVEL_TRACE, "** REQUEST BODY **")
        log.log(LOG_LEVEL_TRACE, request_body)

        log.log(LOG_LEVEL_TRACE, "==============[ BUILDING REQUEST END ]==============")

        try:
            async with self._websession.post(
                url=API_ENDPOINT_URL, data=request_body, headers=request_headers
            ) as resp:

                # We can't use resp.json() because server returns JSON object in response with incorrect mimetype. This causes aiohttp to raise an aiohttp.client_exceptions.ContentTypeError exception.
                raw_response = await resp.text()
        except (
            asyncio.TimeoutError,
            aiohttp.ClientError,
            asyncio.exceptions.CancelledError,
        ) as err:
            log.error("Failed to send request.")

            raise CommunicationError from err

        log.log(LOG_LEVEL_TRACE, "RAW SERVER RESPONSE:\n%s\n\n", raw_response)

        #
        # Validate response format.
        #

        # Expected response format is {"Response": PACKED_RESPONSE_CONTENT}

        try:
            raw_response_json = dict(json.loads(raw_response))
        except (json.JSONDecodeError, TypeError) as err:
            raise ResponseFormatError("Server response not a JSON dict.") from err

        # If auth token is returned, update local.
        if auth_token := resp.headers.get(AUTH_TOKEN_KEY):
            self._auth_token = auth_token

        # Isolate response content

        if not (response_content := raw_response_json.get("Response")):
            raise UnexpectedError("Couldn't find response content.")

        # Unpack response

        unpacked_content = MessagePacker.unpack_server_response(response_content)

        if not unpacked_content:
            raise UnexpectedError("Missing unpacked content.")

        log.log(LOG_LEVEL_TRACE, "UNPACKED RESPONSE CONTENT:\n%s\n\n", unpacked_content)

        response_code = unpacked_content.get(RESULT_CODE_KEY)

        # Check response code (in body).

        if (
            not response_code
            or (
                (response_code := unpacked_content.get(RESULT_CODE_KEY))
                == ServerResponseCodes.INVALID_REQUEST
            )
            or (response_code == ServerResponseCodes.INPUT_MALFORMED and no_retry)
        ):
            log.debug("UNPACKED RESPONSE CONTENT:\n%s\n\n", unpacked_content)
            raise Rejected

        if response_code == ServerResponseCodes.INPUT_MALFORMED:
            # This error may occur if the server doesn't like the submitted first_request_id. We should retry once after clearing tokens and logging back in.

            self._first_request_id = None
            self._auth_token = EMPTY_AUTH_TOKEN

            try:
                if not self._username or not self._password:
                    raise AuthenticationError

                await self.async_login(username=self._username, password=self._password)
            except Exception as err:
                raise Rejected("Request failed even after re-trying login.") from err

            return await self._send_request(request_json=request_json, no_retry=True)

        if response_code == ServerResponseCodes.INVALID_CREDENTIALS:
            raise AuthenticationError

        if response_code == ServerResponseCodes.TRY_AGAIN_LATER_BAD_REQUEST:
            raise CommunicationError

        if response_code == ServerResponseCodes.TRY_AGAIN_LATER_SWIPE_FAILED:
            log.error(msg := "Swipe failed. Machine is probably offline.")
            raise VendFailure(msg)

        if response_code != ServerResponseCodes.SUCCESS:
            log.error(
                "Got unexpected response code %s (%s).",
                response_code,
                unpacked_content.get(RESULT_TEXT_KEY),
            )
            raise UnexpectedError

        log.log(
            LOG_LEVEL_TRACE, "EXTRACTED RESPONSE CONTENT:\n%s\n\n", unpacked_content
        )

        return unpacked_content



"""Functions for packing and unpacking messages."""

from __future__ import annotations

import base64
import gzip
import json
import logging
import urllib.parse
import uuid

from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import modes

from .const import AES_IV
from .const import AES_SUFFIX_PREAUTH
from .const import LOG_LEVEL_TRACE
from .exceptions import MessagePackerError

log = logging.getLogger(__name__)


class MessagePacker:
    """Functions for packing and unpacking client <-> server messages."""

    @staticmethod
    def unpack_server_response(response_body: bytes) -> dict | None:
        """Unpack API response into string."""

        # Unpacking Steps:
        #     1. Base64 decode.
        #     2. gzip decompress.

        log.log(
            LOG_LEVEL_TRACE, "==============[ UNPACKING RESPONSE BEGIN ]=============="
        )

        decoded_as_bytes = base64.standard_b64decode(response_body)

        log.log(
            LOG_LEVEL_TRACE,
            "Base64Decode Bytes -> Bytes:\n%s\n\n",
            decoded_as_bytes.hex(),
        )

        gunzipped_as_bytes = gzip.decompress(decoded_as_bytes)

        log.log(
            LOG_LEVEL_TRACE, "gunzip Bytes -> Bytes:\n%s\n\n", decoded_as_bytes.hex()
        )

        try:
            json_response = json.loads(str(gunzipped_as_bytes, "utf-8"))
        except ValueError as err:
            raise err

        log.log(LOG_LEVEL_TRACE, "UNPACKED RESPONSE:\n%s\n\n", json_response)

        log.log(
            LOG_LEVEL_TRACE, "==============[ UNPACKING RESPONSE END ]=============="
        )

        return json_response if isinstance(json_response, dict) else None

    @staticmethod
    def pack_client_request(
        request_body: str | list | dict,
        first_request_id: str | None = None,
        new_request_id: str | None = None,
    ) -> tuple[str, str]:
        """Pack request body for transmission to server."""

        # https://gist.github.com/brysontyrrell/7cebfb05105c25d00e84ed35bd821dfe

        # new_request_id generated automatically. If debugging intercepted communications, provide actual new_request_id

        # Packing Steps:
        #     1. Apply PKCS #7 padding.
        #     2. Encrypt using 128-bit AES/CBC.
        #     3. Base64 encode.
        #     4. Base64 encode, again.
        #     5. URL encode.

        log.log(
            LOG_LEVEL_TRACE, "==============[ PACKING REQUEST BEGIN ]=============="
        )

        log.log(LOG_LEVEL_TRACE, "Original Request Body:\n%s\n\n", request_body)

        if not new_request_id:
            new_request_id = str(uuid.uuid4())

        log.log(LOG_LEVEL_TRACE, "New Request ID: %s", new_request_id)

        key = MessagePacker._generate_aes_key(
            new_request_id=new_request_id, first_request_id=first_request_id
        )

        # PKCS#7 Pad
        padded_request = MessagePacker._pkcs7_pad(bytes(str(request_body), "utf-8"))

        log.log(LOG_LEVEL_TRACE, "Padded Request:\n%s\n\n", padded_request)

        # AES Encrypt
        cipher = Cipher(algorithms.AES(key), modes.CBC(AES_IV))
        encryptor = cipher.encryptor()
        encrypted_request = (
            encryptor.update(bytes(padded_request)) + encryptor.finalize()
        )

        log.log(
            LOG_LEVEL_TRACE,
            "Encrypted Request:\n%s\n\n",
            MessagePacker._format_hex(encrypted_request),
        )

        # 2x Base 64
        b64_encoded_request = base64.urlsafe_b64encode(
            base64.b64encode(encrypted_request)
        )

        log.log(LOG_LEVEL_TRACE, "Base64 Encoded Request:\n%s\n\n", b64_encoded_request)

        # URL Encode
        url_encoded_request = urllib.parse.quote(b64_encoded_request)

        log.log(LOG_LEVEL_TRACE, "URL Encoded Request:\n%s\n\n", url_encoded_request)

        log.log(LOG_LEVEL_TRACE, "==============[ PACKING REQUEST END ]==============")

        return (str(new_request_id), url_encoded_request)

    @staticmethod
    def unpack_client_request(
        request_body: str,
        new_request_id: str,
        first_request_id: str | None = None,
    ) -> str:
        """Unack request body for message transmitted to server. This helps debug manually intercepted communications and would not typically be used when building an interface for this API."""

        # Packing Steps: Reverse of pack_client_request.

        key = MessagePacker._generate_aes_key(
            new_request_id=new_request_id, first_request_id=first_request_id
        )

        # URL Decode
        url_decoded_request = urllib.parse.unquote(request_body)

        log.log(
            LOG_LEVEL_TRACE,
            "[unpack_client_request] URL Decoded Request:\n%s\n\n",
            url_decoded_request,
        )

        # 2x Base 64 Decode
        b64_decoded_request = base64.b64decode(base64.b64decode(url_decoded_request))

        log.log(
            LOG_LEVEL_TRACE,
            "[unpack_client_request] First Base64 Decoded Request:\n%s\n\n",
            base64.b64decode(url_decoded_request),
        )

        log.log(
            LOG_LEVEL_TRACE,
            "[unpack_client_request] Second Base64 Decoded Request:\n%s\n\n",
            MessagePacker._format_hex(b64_decoded_request),
        )

        # AES Decrypt
        cipher = Cipher(algorithms.AES(key), modes.CBC(AES_IV))
        decryptor = cipher.decryptor()
        decrypted_request = decryptor.update(b64_decoded_request) + decryptor.finalize()

        log.log(
            LOG_LEVEL_TRACE,
            "[unpack_client_request] Decrypted Request:\n%s\n\n",
            decrypted_request.hex(),
        )

        # Unpad PKCS#7
        try:
            unpadded_request = MessagePacker._pkcs7_unpad(decrypted_request)
        except UnicodeDecodeError as err:
            raise MessagePackerError("Error unpadding message.") from err

        log.log(
            LOG_LEVEL_TRACE,
            "[unpack_client_request] Unpadded Request:\n%s\n\n",
            unpadded_request,
        )

        return str(unpadded_request, "utf-8")

    @staticmethod
    def _generate_aes_key(
        new_request_id: str, first_request_id: str | None = None
    ) -> bytes:
        """Generate AES encryption key."""

        key_str = ""
        key_bytes = b""

        if not first_request_id:

            #
            # Generate key for authentication request. Authentication requests identified by missing first_request_id.
            #

            key_base = ""

            length = len(new_request_id) - 1
            while length >= 0 and len(key_base) < 8:
                key_base = key_base + str(new_request_id[length])
                length -= 2

            key_str = key_base + AES_SUFFIX_PREAUTH
            key_bytes = bytes(key_str, "utf-8")

        else:

            # Generate key for all non-authentication requests by weaving together backwards last and second to last request IDs.

            new_req_id_length = len(str(new_request_id)) - 1
            while new_req_id_length >= 0 and len(key_str) < 8:
                key_str = key_str + str(new_request_id)[new_req_id_length]
                new_req_id_length -= 2

            original_uuid_length = len(str(first_request_id)) - 1
            while original_uuid_length >= 0 and len(key_str) < 16:
                key_str = key_str + str(first_request_id)[original_uuid_length]
                original_uuid_length -= 2

            key_bytes = bytes(key_str, "utf-8")

        log.log(
            LOG_LEVEL_TRACE,
            "Encryption Key: %s (%s)",
            key_str,
            MessagePacker._format_hex(key_bytes),
        )

        return key_bytes

    @staticmethod
    def _pkcs7_pad(message: bytes) -> bytes:
        """Add PKCS #7 padding to a string."""
        return message + bytes(
            chr(16 - len(message) % 16) * (16 - len(message) % 16), "utf-8"
        )

    @staticmethod
    def _pkcs7_unpad(message: bytes) -> bytes:
        """Strip PKCS #7 padding from a string."""
        return message[: -message[-1]]

    @staticmethod
    def _format_hex(hex_input: bytes) -> str:
        """Format hex number grouped by bytes."""

        return hex_input.hex(" ").upper()



"""pylaundry constants."""

from enum import IntEnum

#
# Requests
#

API_ENDPOINT_URL = "https://mapp.mylaundrylink.com/AppRequestHandler.aspx"
AUTH_TOKEN_KEY = "CP_AUTH_TOKEN"  # nosec
RESULT_CODE_KEY = "ResultCode"
RESULT_TEXT_KEY = "ResultText"
REFRESH_REQUEST_PREHASH_SUFFIX = (  # Appended to user ID, then MD5 hashed. Used in body data refresh requests.
    "b1c/B?D(G+1bPeSh"
)
EMPTY_AUTH_TOKEN = "00000000-0000-0000-0000-000000000000"  # nosec
APPKEY = "$#!@ES(*#D3$!318z"


class VendResultCodes(IntEnum):
    """Enum of known vend response codes."""

    SUCCESS = 161


class ServerResponseCodes(IntEnum):
    """Enum of known server response codes."""

    INPUT_MALFORMED = -1
    PARENT_OBJ_ERROR = 0
    SUCCESS = 1
    INVALID_CREDENTIALS = 105
    INVALID_USER_ID = 108
    TRY_AGAIN_LATER_BAD_REQUEST = 110  # Caused by incorrect request parameter.
    TRY_AGAIN_LATER_SWIPE_FAILED = 118  # Machine is probably offline.
    INVALID_REQUEST = 122


#
# Encryption
#

AES_IV = bytearray(
    [83, 71, 26, 58, 54, 35, 22, 11, 83, 71, 26, 58, 54, 35, 22, 11]
)  # Used an initialization vector for all AES operations. As hex: ['53', '47', '1A', '3A', '36', '23', '16', '0B', '53', '47', '1A', '3A', '36', '23', '16', '0B']
AES_SUFFIX_PREAUTH = (  # Appended to random request UUID before further processing into AES key. Used for authentication requests only.
    "R&%76mhK"
)

LOG_LEVEL_TRACE = 5



"""HTTP request/response bodies for tests."""

import base64
import gzip
from importlib import resources
import json


def pack_server_response(response_body: str) -> bytes:
    """Pack API response into string to simulate actual server response."""

    response_bytes = bytes(response_body, "utf-8")

    gzipped_as_bytes = gzip.compress(response_bytes)

    encoded_as_bytes = str(base64.standard_b64encode(gzipped_as_bytes), "utf-8")

    return bytes(json.dumps({"Response": encoded_as_bytes}), "utf-8")


def get_http_body(name: str) -> bytes:
    """Get server/client response/request body from JSON file."""

    return pack_server_response(resources.read_text(__package__, f"{name}.json"))



"""Tests for basic functions."""

# pylint: disable=protected-access, missing-class-docstring, no-self-use

import uuid

import aiohttp
from pylaundry import Laundry
from pylaundry import LaundryMachine
from pylaundry.const import EMPTY_AUTH_TOKEN
from pylaundry.exceptions import AuthenticationError
import pytest

# import json

# from deepdiff import DeepDiff


def test_property__initial_state(laundry: Laundry) -> None:
    """Ensure that login data is ingested correctly."""

    assert not laundry._password
    assert not laundry._username
    assert not laundry._first_request_id

    with pytest.raises(AttributeError):
        assert getattr(laundry, "encryption_keys")

    with pytest.raises(AttributeError):
        assert getattr(laundry, "profile")

    with pytest.raises(AttributeError):
        assert getattr(laundry, "machines")

    assert isinstance(laundry._websession, aiohttp.ClientSession)

    assert uuid.UUID(laundry.installation_token)

    assert laundry._auth_token == EMPTY_AUTH_TOKEN


@pytest.mark.asyncio  # type: ignore
async def test__async_login__success__basic_info(
    laundry: Laundry, authentication__response__success: pytest.fixture
) -> None:
    """Test authentication success."""

    await laundry.async_login(username="test@example.com", password="hunter2")

    assert laundry._username == "test@example.com"
    assert laundry._password == "hunter2"

    assert uuid.UUID(laundry._first_request_id)
    assert uuid.UUID(laundry._auth_token)
    assert uuid.UUID(laundry.installation_token)
    assert laundry.machines
    assert laundry.profile


@pytest.mark.asyncio  # type: ignore
async def test__async_login__response__incorrect_credentials(
    laundry: Laundry, authentication__response__incorrect_credentials: pytest.fixture
) -> None:
    """Test authentication failure."""

    with pytest.raises(AuthenticationError):
        await laundry.async_login(
            username="incorrect@example.com", password="incorrect"
        )


@pytest.mark.asyncio  # type: ignore
async def test__async_login__success__profile(
    laundry: Laundry, authentication__response__success: pytest.fixture
) -> None:
    """Test authentication success."""

    await laundry.async_login(username="test@example.com", password="hunter2")

    assert laundry.profile.card_balance == 1.75
    assert (
        laundry.profile.location_address
        == "270 Commerce Dr., Fort Washington, PA 19034"
    )
    assert laundry.profile.location_id == "dd240cf3-7702-5101-8b5b-6dbcdda95d07"
    assert laundry.profile.user_id == "4e353d4d-a9c9-5867-9324-99dbe26d9c35"


@pytest.mark.asyncio  # type: ignore
async def test__async_login__success__machines(
    laundry: Laundry, authentication__response__success: pytest.fixture
) -> None:
    """Test authentication success."""

    # Login
    await laundry.async_login(username="test@example.com", password="hunter2")

    assert isinstance(laundry.machines, dict)

    assert isinstance(
        test_machine := laundry.machines.get("a312b4b7-5110-5775-9966-ed9a6e087e3a"),
        LaundryMachine,
    )

    assert test_machine.minutes_remaining == 0
    assert test_machine.number == "03"
    assert test_machine.busy is False
    assert test_machine.online is True
    assert test_machine.base_price == 1.5


@pytest.mark.asyncio  # type: ignore
async def test__consolidated_refresh__succcess(
    laundry: Laundry,
    authentication__response__success: pytest.fixture,
    consolidated_refresh__response__success: pytest.fixture,
) -> None:
    """Test that consolidated refresh updates macine states."""

    # Login
    await laundry.async_login(username="test@example.com", password="hunter2")

    # Consolidated refresh
    await laundry.async_refresh()

    assert isinstance(laundry.machines, dict)

    assert isinstance(
        test_machine := laundry.machines.get("a312b4b7-5110-5775-9966-ed9a6e087e3a"),
        LaundryMachine,
    )

    assert test_machine.base_price == 200


@pytest.mark.asyncio  # type: ignore
async def test__virtual_vend_topoff__success(
    laundry: Laundry,
    authentication__response__success: pytest.fixture,
    virtual_vend_topoff__response__success: pytest.fixture,
    vend_log_topoff__response__success: pytest.fixture,
) -> None:
    """Test that vend function doesn't raise exception."""

    # Login
    await laundry.async_login(username="test@example.com", password="hunter2")

    # Vend Machine
    await laundry.async_vend("a312b4b7-5110-5775-9966-ed9a6e087e3a")


@pytest.mark.asyncio  # type: ignore
async def test__vend_log_topoff__success(
    laundry: Laundry,
    authentication__response__success: pytest.fixture,
    vend_log_topoff__response__success: pytest.fixture,
) -> None:
    """Test that vend log function doesn't raise exception."""

    # Login
    await laundry.async_login(username="test@example.com", password="hunter2")

    # Log Vend
    await laundry._async_log_vend("a312b4b7-5110-5775-9966-ed9a6e087e3a", 1, True)
